@@ -12,11 +12,11 @@ __license__ = "mit"
 class TestIdentityInitialCovariance(unittest.TestCase):
 
     def setUp(self):
-        self.A = np.eye(2)
-        self.C = np.eye(2)
+        self.A = np.random.randn(2,2)
+        self.I = np.eye(2)
 
     def test_that_R_inverse_is_zero(self):
-        assert np.linalg.norm(mdf.makeRi(self.A, self.C)) < 1E-14
+        assert np.linalg.norm(mdf.makeRi(self.A, self.I)) < 1E-8
 
     def test_solutions_with_orthogonal_map(self):
         # Arrange
@@ -24,7 +24,7 @@ class TestIdentityInitialCovariance(unittest.TestCase):
         A = np.random.randn(2,2)
         b = np.random.randn(2,1)
         # m = np.zeros(2,1)
-        c = self.C
+        c = self.I
 
         # Act
         y = A@t + b
@@ -32,3 +32,7 @@ class TestIdentityInitialCovariance(unittest.TestCase):
         err = sol - t
         # Assert
         assert np.linalg.norm(err) < 1E-8
+
+    def test_updated_cov_has_R_equal_zero_for_full_rank_A(self):
+        up_cov = mdf.updated_cov(self.A, self.I, self.I)
+        assert np.linalg.norm(up_cov - np.linalg.inv(self.A.T@self.A)) < 1E-8
